@@ -1,8 +1,8 @@
 ;;; Dezyne --- Dezyne command line tools
 ;;;
 ;;; Copyright © 2021 Paul Hoogendijk <paul@dezyne.org>
-;;; Copyright © 2021, 2022 Jan (janneke) Nieuwenhuizen <janneke@gnu.org>
 ;;; Copyright © 2020, 2021, 2022, 2023 Rutger van Beusekom <rutger@dezyne.org>
+;;; Copyright © 2021, 2022, 2023 Jan (janneke) Nieuwenhuizen <janneke@gnu.org>
 ;;;
 ;;; This file is part of Dezyne.
 ;;;
@@ -20,29 +20,27 @@
 ;;; License along with Dezyne.  If not, see <http://www.gnu.org/licenses/>.
 
 (define-module (dzn explore)
-  #:use-module (ice-9 match)
-
   #:use-module (srfi srfi-1)
   #:use-module (srfi srfi-26)
   #:use-module (srfi srfi-71)
 
+  #:use-module (ice-9 match)
   #:use-module (json)
 
-  #:use-module (dzn command-line)
-
-  #:use-module ((oop goops) #:renamer (lambda (x) (if (member x '(<port> <foreign>)) (symbol-append 'goops: x) x)))
-  #:use-module (dzn goops)
+  #:use-module (dzn ast display)
+  #:use-module (dzn ast goops)
   #:use-module (dzn ast)
+  #:use-module (dzn command-line)
   #:use-module (dzn config)
-  #:use-module (dzn display)
   #:use-module (dzn misc)
-  #:use-module (dzn vm ast)
+  #:use-module (dzn vm compliance)
   #:use-module (dzn vm goops)
-  #:use-module (dzn vm runtime)
+  #:use-module (dzn vm normalize)
   #:use-module (dzn vm report)
   #:use-module (dzn vm run)
-  #:use-module (dzn vm step)
+  #:use-module (dzn vm runtime)
   #:use-module (dzn vm util)
+
   #:export (lts
             pc->rtc-lts
             pc->state-number
@@ -56,17 +54,14 @@
 
 (define-method (process-external (pc <program-counter>))
   (let ((queues (.external-q pc)))
-    (if (or (null? queues)
-            (pair? (.async pc))) '()
-            (let ((instances (map car queues)))
-              (append-map (cute run-external-q pc <>) instances)))))
+    (if (null? queues) '()
+        (let ((instances (map car queues)))
+          (append-map (cute run-external-q pc <>) instances)))))
 
 (define (run-to-completion** pc event)
   (let* ((pc (clone pc #:instance #f #:trail '()))
          (pc (reset-replies pc)))
     (cond
-     ((eq? event 'async)
-      (flush-async pc '()))
      ((eq? event 'defer)
       (if (null? (.defer pc)) '()
           (run-defer-event pc event)))
@@ -75,22 +70,16 @@
      ((is-a? (%sut) <runtime:port>)
       (run-to-completion pc event))
      ((requires-trigger? event)
-      (if (pair? (.async pc)) '()
-          (let* ((traces (run-requires pc event))
-                 (stop flush
-                       (partition
-                        (disjoin (compose null? .async car)
-                                 did-provides-out?
-                                 (compose .status car))
-                        traces))
-                 (port (.port (string->trigger event)))
-                 (blocked-port (and=> (blocked-on-boundary? pc) .ast)))
-            (if (and port (ast:eq? port blocked-port)) '()
-                (append stop (append-map (cute flush-async-trace <> '())
-                                         flush))))))
+      (let* ((traces (run-requires pc event))
+             (port (.port (string->trigger event)))
+             (blocked-port (and=> (blocked-on-boundary? pc) .ast)))
+        (if (and port (ast:eq? port blocked-port)) '()
+            traces)))
      ((provides-trigger? event)
       (if (blocked-on-boundary-provides? pc event) '()
-          (run-to-completion pc event))))))
+          (run-to-completion pc event)))
+     (else
+      '()))))
 
 ;; table is initially (<"<illegal>") . 0>)
 (define* ((pc->state-number table count) pc #:optional trail)
@@ -135,69 +124,135 @@ that PC has one more collaterally blocked coroutine on the same port."
       (_
        #f)))
 
-  (define (run-label orig-pc label)
-    (let* ((orig-pc (clone orig-pc #:collateral-instance #f))
-           (sw-pc (switch-context orig-pc))
-           (bob-pc (blocked-on-boundary-switch-context orig-pc)))
-      (if (and (eq? orig-pc sw-pc)
-               (eq? orig-pc bob-pc))
-          (run-to-completion** orig-pc label)
-          (cond
-           ((and (not (eq? orig-pc sw-pc))
-                 (not (provides-trigger? label)))
-            (run-to-completion sw-pc 'rtc))
-           (else
-            (append (let* ((traces (run-to-completion bob-pc 'rtc))
-                           (flush traces
-                                  (partition (compose .running-defer? car)
-                                             traces)))
-                      (append traces
-                              (append-map run-flush flush)))
-                    (run-to-completion** orig-pc label)))))))
-
-  (define (run-labels pc labels)
-    (let loop ((labels labels) (traces '()))
-      (if (null? labels) traces
-          (let* ((label (car labels))
-                 (new (run-label pc label))
-                 (new (if (eq? label 'external) new
-                          (filter-implicit-illegal-only new)))
-                 (new (filter (compose not (is-status? <blocked-error>) car)
-                              new))
-                 (new (filter (compose not (cute collateral? <> pc) car) new))
-                 (new (map (cute rewrite-trace-head prune-defer <>) new)))
-            (cond
-             ((find (compose (is-status? <livelock-error>) car) new)
-              (format (current-error-port) "warning: livelock, bailing out\n")
-              '())
-             (else
-              (loop (cdr labels) (append new traces))))))))
-
-  (let* ((labels (labels))
-         (labels (if (is-a? (%sut) <runtime:port>) labels
-                     (cons* 'async 'defer 'external labels)))
-         (lts (make-hash-table))
+  (let* ((lts (make-hash-table))
          (state-number-table (make-hash-table))
          (state-number-count (list 1))
          (pc->state-number (pc->state-number state-number-table
                                              state-number-count)))
+
+    (define (blocked-prefix pc)
+      (let* ((blocking-pc (rtc-block-pc (cdar (.blocked pc))))
+             (blocking-pc (.previous blocking-pc))
+             (blocked-from (and=> blocking-pc pc->state-number))
+             (pc+traces (hash-ref lts blocked-from)))
+        (match pc+traces
+          ((pc . traces)
+           (or (find (compose pair? .blocked car) traces)
+               '()))
+          (_
+           '()))))
+
+    (define (check-compliance-blocked trace)
+      (let ((orig-pc (last trace)))
+        (let* ((blocked-prefix (blocked-prefix orig-pc))
+               (q-trigger (and (null? blocked-prefix)
+                               (and=>
+                                (as (.q (get-state orig-pc (%sut))) <pair>)
+                                car)))
+               (full-trace (append trace blocked-prefix))
+               (orig-pc (last full-trace))
+               (trigger (or
+                         q-trigger
+                         (any
+                          (conjoin
+                           (compose (is? <runtime:component>) .instance)
+                           .trigger)
+                          (reverse full-trace))))
+               (label (and trigger (trigger->string trigger)))
+               (checked (if (not label) '()
+                            (check-provides-compliance orig-pc label
+                                                       full-trace)))
+               (illegal? (is-status? <implicit-illegal-error>))
+               (checked (filter (compose not illegal? car) checked))
+               (pcs (map car checked))
+               (corrected (if (not (= 1 (length pcs))) trace
+                              (cons (car pcs) trace))))
+          corrected)))
+
+    (define (run-label orig-pc label)
+      (%debug "run-label ~a\n" label)
+      (let* ((orig-pc (clone orig-pc #:collateral-instance #f))
+             (defer? (eq? label 'defer))
+             (sw-pc (if defer? orig-pc
+                        (switch-context orig-pc)))
+             (bob-pc (if defer? orig-pc
+                         (blocked-on-boundary-switch-context orig-pc)))
+             (traces (cond
+                      ((and (eq? orig-pc sw-pc)
+                            (eq? orig-pc bob-pc))
+                       (run-to-completion** orig-pc label))
+                      ((and (not (eq? orig-pc sw-pc))
+                            (not (provides-trigger? label)))
+                       (run-to-completion sw-pc 'rtc))
+                      (else
+                       (let* ((traces (run-to-completion bob-pc 'rtc))
+                              (flush traces
+                                     (partition (compose .running-defer? car)
+                                                traces))
+                              (traces (append traces
+                                              (append-map run-flush flush))))
+                         (append traces
+                                 (run-to-completion** orig-pc label))))))
+             (blocked? (not (eq? bob-pc orig-pc)))
+             (skip-compliance? (or (.skip-compliance? orig-pc)
+                                   (is-a? (%sut) <runtime:port>)
+                                   (null? traces)))
+             (traces (map (cute append <> (list orig-pc)) traces))
+             (traces (cond
+                      ((and (not skip-compliance?) (not blocked?))
+                       (let* ((cpc orig-pc)
+                              (cpc (reset-replies cpc))
+                              (cpc (clone cpc #:instance #f)))
+                         (check-provides-compliance* cpc label traces)))
+                      ((and blocked?
+                            (not (.skip-compliance? orig-pc))
+                            (not (eq? label 'defer))
+                            (not (or (.status pc)
+                                     (is-a? (%sut) <runtime:port>)
+                                     (is-a? (%sut) <runtime:system>))))
+                       (map check-compliance-blocked traces))
+                      (else
+                       traces)))
+             (traces (if (eq? label 'external) traces
+                         (filter-implicit-illegal-only traces)))
+             (traces (filter (compose not (is-status? <blocked-error>) car)
+                             traces))
+             (traces (filter (compose not (cute collateral? <> pc) car) traces)))
+        (map (cute rewrite-trace-head prune-defer <>) traces)))
+
+    (define (run-labels pc labels)
+      (%debug "run-labels ~a\n" labels)
+      (let loop ((labels labels) (traces '()))
+        (if (null? labels) traces
+            (let* ((label (car labels))
+                   (new (run-label pc label)))
+              (cond
+               ((find (compose (is-status? <livelock-error>) car) new)
+                (format (current-error-port) "warning: livelock, bailing out\n")
+                '())
+               (else
+                (loop (cdr labels) (append new traces))))))))
+
     (hash-set! state-number-table "<illegal>" 0)
-    (when (.status pc)
-      (let ((pc0 (clone pc #:status #f)))
-        (hash-set! lts 1 (cons pc0 (list (list pc0 pc))))))
     (let loop ((pc pc))
       (let ((from (pc->state-number pc)))
+        (%debug "loop ~a~a\n" from (if (hash-ref lts from) ": seen!" ""))
         (unless (or (.status pc) (hash-ref lts from))
-          (let* ((from-pc pc)
-                 (traces (run-labels pc labels))
-                 (traces (map (cute append <> (list from-pc)) traces))
-                 (pcs (map car traces)))
-            (hash-set! lts from (cons from-pc traces))
-            (let* ((traces (filter (negate trace-done?) traces))
-                   (pcs (map car traces)))
-              (for-each loop pcs))))))
-    (when (= (car state-number-count) 1)
-      (hash-set! lts 1 (cons pc (list (list pc)))))
+          (let* ((labels (labels pc))
+                 (labels (if (is-a? (%sut) <runtime:port>) labels
+                             (cons* 'defer 'external labels)))
+                 (traces (run-labels pc labels)))
+            (when (or (is-a? (%sut) <runtime:port>)
+                      (any pair? traces))
+              (hash-set! lts from (cons pc traces))
+              (let* ((traces (filter (negate trace-done?) traces))
+                     (pcs (map car traces)))
+                (for-each loop pcs)))))))
+
+    (when (.status pc) ;; range-member
+      (let ((pc0 (clone pc #:status #f)))
+        (hash-set! lts 1 (cons pc0 (list (list pc0 pc))))))
+
     (values lts pc->state-number state-number-count)))
 
 
@@ -210,7 +265,7 @@ that PC has one more collaterally blocked coroutine on the same port."
    ((from from-label transition to trigger-location)
     ...)
 "
-  (define (trigger-location trace)
+  (define (on-location trace)
     (let ((pc (find (conjoin (compose (is? <on>) .statement)
                              (disjoin (const (is-a? (%sut) <runtime:port>))
                                       (compose (is? <runtime:component>) .instance)))
@@ -237,12 +292,12 @@ that PC has one more collaterally blocked coroutine on the same port."
                                    (equal? a-transition b-transition))))))
            (to (map car to+transition))
            (transition (map cdr to+transition))
-           (trigger-location (map trigger-location traces)))
+           (on-location (map on-location traces)))
       (if (null? to) (list (list from pc #f #f #f))
           (map (cut list from pc <> <> <>)
                transition
                to
-               trigger-location))))
+               on-location))))
 
   (let* ((alist (hash-table->alist lts))
          (alist (sort alist
@@ -503,11 +558,18 @@ RTC-LTS->STATE-DIAGRAM."
             (cons illegal-pc trace)))))
 
   (define (fix-missing-reply-trace trace)
-    (match trace
-      (((and ($ <program-counter>) (? (is-status? <missing-reply-error>)) error)
-        ($ <program-counter>) tail ...)
-       (cons error tail))
-      (_ trace)))
+    (let* ((pc (car trace))
+           (status (.status pc)))
+      (if (not (is-a? status <missing-reply-error>)) trace
+          (let* ((instance
+                  (or (any (compose (is? <runtime:component>) .instance) trace)
+                      (any (compose (is? <runtime:port>) .instance) trace)))
+                 (return (list-index
+                          (conjoin (negate (is-status? <missing-reply-error>))
+                                   (compose (cute eq? <> instance) .instance)
+                                   (compose (is? <trigger-return>) .statement))
+                          trace)))
+            (cons pc (drop trace (1+ return)))))))
 
   (define (trail->state-numbers pc trail to)
     "Use every trail prefix to hash pc and produce contiguous state
@@ -535,7 +597,10 @@ triple per label in the trail of the transition."
 
   (define (transition->aut from pc+traces)
     (let ((pc traces (match pc+traces ((pc . traces) (values pc traces)))))
-      (append-map (cute trace->lts-triples pc from <>) traces)))
+      (delete-duplicates
+       (append-map
+        (cute trace->lts-triples pc from <>)
+        traces))))
 
   (values
    (cons
@@ -556,45 +621,54 @@ RTC-LTS->LTS."
 ;;;
 ;;; Entry points.
 ;;;
-
-(define* (state-diagram root #:key format model
+(define* (state-diagram ast #:key format model
                         queue-size queue-size-defer queue-size-external
                         ports? extended? actions? labels? returns?)
   "Entry-point for dzn explore --state-diagram."
-  (parameterize ((%debug? (> (dzn:debugity) 0))
-                 (%exploring? #t)
-                 (%queue-size queue-size)
-                 (%queue-size-defer queue-size-defer)
-                 (%queue-size-external queue-size-external)
-                 (%sut (runtime:get-sut root model)))
-    (parameterize ((%instances (runtime:create-instances (%sut))))
-      (let* ((pc (make-pc))
-             (lts pc->state-number state-count (pc->rtc-lts pc))
-             (size (1- (car state-count)))
-             (lts pc->state-number state-count
-                  (lts-remove lts size
-                              #:ports? (or ports? extended?)
-                              #:extended? extended?
-                              #:actions? actions?
-                              #:labels? labels?
-                              #:returns? returns?
-                              #:self? (or extended? ports?)))
-             (state-diagram (rtc-lts->state-diagram lts pc->state-number)))
-        (if (equal? format "json") (display
-                                    (state-diagram->json
-                                     state-diagram (.working-directory root)))
-            (display (state-diagram->dot state-diagram (pc->hash pc))))))))
+  (let* ((root (vm:normalize ast))
+         (model (ast:get-model root (ast:dotted-name model))))
+    (when (> (dzn:debugity) 1)
+      (ast:pretty-print root (current-error-port)))
+    (parameterize ((%compliance-check? #f)
+                   (%debug? (> (dzn:debugity) 0))
+                   (%exploring? #t)
+                   (%queue-size queue-size)
+                   (%queue-size-defer queue-size-defer)
+                   (%queue-size-external queue-size-external)
+                   (%sut (runtime:get-sut root model)))
+      (parameterize ((%instances (runtime:create-instances (%sut))))
+        (let* ((pc (make-pc))
+               (lts pc->state-number state-count (pc->rtc-lts pc))
+               (size (1- (car state-count)))
+               (lts pc->state-number state-count
+                    (lts-remove lts size
+                                #:ports? (or ports? extended?)
+                                #:extended? extended?
+                                #:actions? actions?
+                                #:labels? labels?
+                                #:returns? returns?
+                                #:self? (or extended? ports?)))
+               (state-diagram (rtc-lts->state-diagram lts pc->state-number)))
+          (if (equal? format "json") (display
+                                      (state-diagram->json
+                                       state-diagram (.working-directory root)))
+              (display (state-diagram->dot state-diagram (pc->hash pc)))))))))
 
-(define* (lts root #:key model queue-size queue-size-defer queue-size-external)
+(define* (lts ast #:key model queue-size queue-size-defer queue-size-external)
   "Entry-point for dzn explore --lts."
-  (parameterize ((%debug? (> (dzn:debugity) 0))
-                 (%exploring? #t)
-                 (%queue-size queue-size)
-                 (%queue-size-defer queue-size-defer)
-                 (%queue-size-external queue-size-external)
-                 (%sut (runtime:get-sut root model)))
-    (parameterize ((%instances (runtime:create-instances (%sut))))
-      (let* ((pc (make-pc))
-             (rtc-lts pc->state-number state-count (pc->rtc-lts pc))
-             (lts state-count (rtc-lts->lts rtc-lts pc->state-number state-count)))
-        (lts->aut lts state-count)))))
+  (let* ((root (vm:normalize ast))
+         (model (ast:get-model root (ast:dotted-name model))))
+    (when (> (dzn:debugity) 1)
+      (ast:pretty-print root (current-error-port)))
+    (parameterize ((%compliance-check? #f)
+                   (%debug? (> (dzn:debugity) 0))
+                   (%exploring? #t)
+                   (%queue-size queue-size)
+                   (%queue-size-defer queue-size-defer)
+                   (%queue-size-external queue-size-external)
+                   (%sut (runtime:get-sut root model)))
+      (parameterize ((%instances (runtime:create-instances (%sut))))
+        (let* ((pc (make-pc))
+               (rtc-lts pc->state-number state-count (pc->rtc-lts pc))
+               (lts state-count (rtc-lts->lts rtc-lts pc->state-number state-count)))
+          (lts->aut lts state-count))))))
